# Check that we can define a pure python array class.
# An array is a sequence of monotonic data values.
"""
This clas provide an array like object to access a string or any object wich
define the buffer interface.
The wrapped object must also define a .resize(n) method if the emulated
sequence  need to be able to change it's size.
"""

from __future__ import generators

from mem import chunk
from mem import tools
from mem import memslice
import struct

 
class StrArray(object):
    def __init__(self, format, data=None):
        self.typecode = format
        self.itemsize = struct.calcsize(format)
        if data is None:
            self.__data = chunk.chunk()
        else:
            if tools.memlen(data)%self.itemsize:
                raise ValueError("data buffer size doesn't match format")
            self.__data = data

    def __getitem__(self, index):
        data = self.__data
        itemsize = self.itemsize
        if isinstance(index, type(slice(0))):
            start, stop, step = index.indices(len(self))
            if step != 1:
                raise TypeError("sequence index must be integer")
            start*=itemsize
            stop *=itemsize
            try:
                nd = data[start:stop]
            except TypeError:
                nd = memslice.memslice(data, start, stop)
            return StrArray(self.typecode, nd)
        else:
            start = index*itemsize
            if 0<=start<tools.memlen(data):
                return struct.unpack(self.typecode,
                            memslice.memslice(data, start, start+itemsize))[0]
            else:
                raise IndexError, "index out of range"

    def __setitem__(self, index, value):
        data = self.__data
        itemsize = self.itemsize
        if isinstance(index, type(slice(0))):
            start, stop, step = index.indices(len(self))
            if step!=1:
                raise TypeError("sequence index must be integer")
            if isinstance(value, StrArray) and value.typecode==self.typecode:
                nd = value.__data
            else:
                nd = Array(self.typecode, value).__data
            # we avoid slicing on data because it may require something
            # else than a sequence of bytes.
            start *= itemsize
            stop *= itemsize
            length = len(data)
            delta = len(nd) + start -stop
            if delta>0:
                data.resize(length+delta)                
                tools.memcopy(data, memslice.memslice(data, stop, length),
                              stop+delta)
            elif delta<0:
                resize = data.resize
                tools.memcopy(data, memslice.memslice(data, stop), stop+delta)
                resize(length+delta)
            tools.memcopy(data, nd, start)
        else:
            start = index*itemsize
            if 0<=start<tools.memlen(data):
                tools.memcopy(data, struct.pack(self.typecode, value), start)
            else:
                raise IndexError, "index out of range"

    def __delitem__(self, index):
        data = self.__data
        itemsize = self.itemsize
        if isinstance(index, type(slice(0))):
            start, stop, step = index.indices(len(self))
            if step != 1:
                raise TypeError("sequence index must be integer")
        else:
            if 0<=index<len(self):
                start = index
                stop = start+1
            else:
                raise IndexError, "index out of range"
        start *= itemsize
        stop *= itemsize
        length = len(data)
        resize = data.resize
        tools.memcopy(data, memslice.memslice(data, stop), start)
        resize(length+start-stop)
        
    def __len__(self):
        return tools.memlen(self.__data)/self.itemsize

    def __iter__(self):
        data = self.__data
        step = self.itemsize
        for i in range(0, tools.memlen(data), step):
            yield struct.unpack(self.typecode,
                                memslice.memslice(data, i, i+step))[0]

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.typecode,
                               tuple(self))

    def __iadd__(self, other):
        if isinstance(other, StrArray) and other.typecode==self.typecode:
            ds = self.__data
            do = other.__data
            ds.resize(tools.memlen(ds)+tools.memlen(do), do)
        else:
            start = len(self)
            self.__data.resize((start + len(other))*self.itemsize)
            for i in other:
                self[start] = i
                start += 1
        return self
    extend = __iadd__
    def __imul__(self, nb):
        if isinstance(nb, int):
            data = self.__data
            data.resize(max(nb,0)*tools.memlen(data), data)
            return self
        else:
            raise ValueError("unsupported operand type(s) for * :%r and %r" %
                             (type(self).__name__, type(nb).__name__))

    def __cmp__(self, other):
        try:
            it = iter(other)
        except TypeError:
            return cmp(type(self), type(other))
        for i in self:
            try:
                j = it.next()
            except StopIteration:
                return 1
            r = cmp(i,j)
            if r:
                return r
        try:
            j = it.next()
        except StopIteration:
            return 0
        return -1

    def append(self, value):
        data = self.__data
        size = tools.memlen(data)
        data.resize(size+self.itemsize)
        tools.memcopy(data, struct.pack(self.typecode, value), size)

    def index(self, value, start=0, stop=None):
        n = self.itemsize
        motif = struct.pack(self.typecode, value)
        if stop is None: stop=len(self)
        return tools.index(self.__data, motif, start*n, stop*n, n)/n

    def rindex(self, value, start=0, stop=None):
        n = self.itemsize
        motif = struct.pack(self.typecode, value)
        if stop is None: stop=len(self)
        return tools.rindex(self.__data, motif, start*n, stop*n, n)/n

    def count(self, value):
        result = 0
        motif = struct.pack(self.typecode, value)
        n = self.itemsize
        start = 0
        stop=len(self)*n
        while 1:
            try:
                index = tools.index(self.__data, motif, start, stop, n)
            except ValueError:
                return result
            result += 1
            start = index + n

    def insert(self, index, value):
        self[index:index] = (value,)

    def pop(self, index):
        v = self[index]
        del self[index]
        return v

    def remove(self, value):
        try:
            index = self.index(value)
        except ValueError:
            return
        del self[index]

    def reverse(self):
        n = len(self)
        for i in range(n//2):
            self[i],self[n-i-1] = self[n-i-1],self[i]

    def sort(self):
        L = list(self)
        L.sort()
        self[:] = L

class Array(StrArray):
    def __init__(self, format, data=()):
        StrArray.__init__(self, format)
        self += data
    def __mul__(self, nb):
        r = Array(self.typecode, self)
        r *= nb
        return r
    def __add__(self, other):
        r = Array(self.typecode, self)
        r += other
        return r

