/* 
Copyright (c) 2003, Sebastien Keim, <s.keim@laposte.net>.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not
be used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization of the copyright holder.
*/

#include "Python.h"

PyDoc_STRVAR(module__doc__, "raw memory segment.\n\
\n\
This module define a new object types wich can be used to allocate\n\
raw segments of memory.\n\
Theses objects support the single-segment buffer interface.");


typedef struct {
    PyObject_HEAD int c_size;
    char *c_ptr;
} PyChunkObject;

extern PyTypeObject PyChunk_Type;

/* -------------- private tools ------------------*/

static PyChunkObject *chunk_alloc(int size)
{
    PyChunkObject *chunk = PyObject_New(PyChunkObject, &PyChunk_Type);

    if (chunk == NULL)
	return NULL;
    if (size > 0) {
	chunk->c_ptr = malloc(size);
	if (!chunk->c_ptr) {
	    PyErr_NoMemory();
	    Py_DECREF(chunk);
	    return NULL;
	}
	chunk->c_size = size;
    } else {
	chunk->c_size = 0;
	chunk->c_ptr = NULL;
    }
    return chunk;
}

static int chunk_do_resize(PyChunkObject * self, int size, PyObject * data)
{
    self->c_ptr = realloc(self->c_ptr, size);

    if (size) {
	if (!self->c_ptr) {
	    self->c_size = 0;
	    PyErr_NoMemory();
	    return -1;
	}
    } else {
	self->c_ptr = NULL;
    }

    if (size > self->c_size) {
	char *ptr = self->c_ptr + self->c_size;
	char *end = self->c_ptr + size;

	if (data) {
	    const void *data_ptr = NULL;
	    int data_length = 0;

	    /* The data object can be the chunk itself. */
	    if (PyObject_AsReadBuffer(data, &data_ptr, &data_length))
		return -1;
	    if (data_length == 0) {
		PyErr_SetString(PyExc_ValueError, "data can't be empty");
		return -1;
	    }
	    while (ptr + data_length <= end) {
		memmove(ptr, data_ptr, data_length);
		ptr += data_length;
	    }
	    memmove(ptr, data_ptr, end - ptr);
	} else {
	    memset(ptr, 0, end - ptr);
	}
    }
    self->c_size = size;
    return 0;
}

/* chunk object */

static PyObject *chunk_subtype_new(PyTypeObject * type, PyObject * args,
				   PyObject * kw);

static PyObject *chunk_new(PyTypeObject * type, PyObject * args,
			   PyObject * kw)
{
    PyChunkObject *chunk = NULL;
    int size = 0;
    int i;
    void* ptr;

    if (type != &PyChunk_Type)
	return chunk_subtype_new(type, args, kw);

    for (i=0; i<PyTuple_Size(args); i++){
        int  s;
        const void *p;
        if (PyObject_AsReadBuffer(PyTuple_GetItem(args, i), &p, &s))
	    return NULL;
	size += s;
    }
    chunk = chunk_alloc(size);
    if(chunk == NULL)
        goto error;
    ptr = chunk->c_ptr;
    for (i=0; i<PyTuple_Size(args); i++){
        int  s;
        const void *p;
	if (PyObject_AsReadBuffer(PyTuple_GetItem(args, i), &p, &s))
	    goto error;
	if (s > size){
	  PyErr_SetString(PyExc_ValueError, "buffer overflow");
	    goto error;
	}
	memcpy(ptr, p, s);
	size -= s;
	ptr += s;
    }
    if (size>0){
        PyErr_SetString(PyExc_ValueError, "buffer underflow");
        goto error;
    }

    return (PyObject *) chunk;

error:
    Py_XDECREF(chunk);
    return NULL;

}

static PyObject *chunk_subtype_new(PyTypeObject * type, PyObject * args,
				   PyObject * kw)
{
    PyChunkObject *tmp, *new;

    assert(PyType_IsSubtype(type, &PyChunk_Type));
    tmp = (PyChunkObject *) chunk_new(&PyChunk_Type, args, kw);
    if (tmp == NULL)
	return NULL;

    new = (PyChunkObject *) type->tp_alloc(type, 0);
    if (new == NULL)
	return NULL;

    new->c_size = tmp->c_size;
    new->c_ptr = tmp->c_ptr;

    tmp->c_size = 0;		/* we take ownership of the buffer */
    tmp->c_ptr = NULL;		/* so we don't want it to be freed by */
    Py_DECREF(tmp);		/* chunk_dealloc(tmp) */
    return (PyObject *) new;
}

static void chunk_dealloc(PyChunkObject * chunk)
{
    if (chunk->c_ptr) {
	free(chunk->c_ptr);
	chunk->c_ptr = NULL;
    }
    chunk->ob_type->tp_free((PyObject *) chunk);
}

static PyObject *chunk_str(PyChunkObject * self)
{
    return PyString_FromStringAndSize(self->c_ptr, self->c_size);
}

static PyObject *chunk_repr(PyChunkObject * self)
{

    PyObject *result = NULL;
    PyObject *data = NULL;
    PyObject *repr = NULL;

    data = PyString_FromStringAndSize(self->c_ptr, self->c_size);
    if (!data)
	goto end;
    repr = PyObject_Repr(data);
    if (!repr)
	goto end;

    result = PyString_FromFormat("%s(%s)",
				 self->ob_type->tp_name,
				 PyString_AsString(repr));
  end:
    Py_XDECREF(data);
    Py_XDECREF(repr);
    return result;
}



static PyObject *chunk_richcompare(PyObject * va, PyObject * vb, int op)
{
    PyChunkObject *a = (PyChunkObject *) va;
    PyChunkObject *b = (PyChunkObject *) vb;
    int c;
    int len_a, len_b;
    int min_len;
    PyObject *result;

    /* Make sure both arguments are chunks. */
    if (!(PyObject_TypeCheck(a, &PyChunk_Type) &&
	  PyObject_TypeCheck(b, &PyChunk_Type))) {
	result = Py_NotImplemented;
	goto out;
    }
    if (a == b) {
	switch (op) {
	case Py_EQ:
	case Py_LE:
	case Py_GE:
	    result = Py_True;
	    goto out;
	case Py_NE:
	case Py_LT:
	case Py_GT:
	    result = Py_False;
	    goto out;
	}
    }
    if (op == Py_EQ) {
	/* Supporting Py_NE here as well does not save
	   much time, since Py_NE is rarely used.  */
	if ((a->c_size == b->c_size)
	    && (memcmp(a->c_ptr, b->c_ptr, a->c_size) == 0)) {
	    result = Py_True;
	} else {
	    result = Py_False;
	}
	goto out;
    }
    len_a = a->c_size;
    len_b = b->c_size;
    min_len = (len_a < len_b) ? len_a : len_b;
    if (min_len > 0) {
	c = memcmp(a->c_ptr, b->c_ptr, min_len);
    } else
	c = 0;
    if (c == 0)
	c = (len_a < len_b) ? -1 : (len_a > len_b) ? 1 : 0;
    switch (op) {
    case Py_LT:
	c = c < 0;
	break;
    case Py_LE:
	c = c <= 0;
	break;
    case Py_EQ:
	assert(0);
	break;			/* unreachable */
    case Py_NE:
	c = c != 0;
	break;
    case Py_GT:
	c = c > 0;
	break;
    case Py_GE:
	c = c >= 0;
	break;
    default:
	result = Py_NotImplemented;
	goto out;
    }
    result = c ? Py_True : Py_False;
  out:
    Py_INCREF(result);
    return result;
}


/* ------------------------- Pickling support ------------------------- */
/* __getstate__ isn't exposed */
static PyObject* 
chunk_getstate(PyChunkObject * self)
{
    return Py_BuildValue("(N)",
		 PyString_FromStringAndSize(self->c_ptr, self->c_size));
}

PyDoc_STRVAR(reduce__doc__, "c.__reduce__() -->  (cls, state)\n\
\n\
pickle and copy support.\n");
static PyObject* 
chunk_reduce(PyChunkObject * self)
{
    return Py_BuildValue("ON", self->ob_type, chunk_getstate(self));
}


/* ------------------- sequence protocol ----------------------------*/


static void normalize_slice(int size, int *start, int *end)
{
    if (*start > size)
	*start = size;
    if (*end < *start)
	*end = *start;
    else if (*end > size)
	*end = size;
}

static int chunk_length(PyChunkObject * self)
{
    return self->c_size;
}

static PyObject *chunk_item(PyChunkObject * a, int i)
{
    if (i < 0 || i >= a->c_size) {
	PyErr_SetString(PyExc_IndexError, "index out of range");
	return NULL;
    }
    return PyString_FromStringAndSize(a->c_ptr + i, 1);
}

static PyObject *chunk_slice(PyChunkObject * a, int start, int stop)
{
    PyChunkObject *np;
    
    normalize_slice(a->c_size, &start, &stop);

    np = chunk_alloc(stop - start);
    if (!np)
	return NULL;

    memcpy(np->c_ptr, a->c_ptr + start, stop - start);

    return (PyObject *) np;
}


static PyObject *chunk_concat(PyChunkObject * a, PyObject * b)
{
    int size;
    const void *ptr;
    PyChunkObject *np;
    if (PyObject_AsReadBuffer(b, &ptr, &size))
	return NULL;

    np = chunk_alloc(a->c_size + size);
    if (!np)
	return NULL;

    memcpy(np->c_ptr, a->c_ptr, a->c_size);
    memcpy(np->c_ptr + a->c_size, ptr, size);

    return (PyObject *) np;
}

static PyObject *chunk_repeat(PyChunkObject * a, int n)
{
    PyChunkObject *np;
    int i;

    if (n < 0)
	n = 0;
    np = chunk_alloc(a->c_size * n);
    if (!np)
	return NULL;

    for (i = 0; i < n; i++) {
	memcpy(np->c_ptr + i * a->c_size, a->c_ptr, a->c_size);
    }

    return (PyObject *) np;
}

static int chunk_ass_item(PyChunkObject * a, int i, PyObject * v)
{
    int size;
    const void *ptr;

    if (i < 0 || i >= a->c_size) {
	PyErr_SetString(PyExc_IndexError,
			"chunk assignment index out of range");
	return -1;
    }
    if(v==NULL) {  /* __del_item__ */
      	memcpy(a->c_ptr+i, a->c_ptr+i+1, a->c_size-1);
	a->c_size -= 1;
	a->c_ptr = realloc(a->c_ptr, a->c_size);
	if((a->c_ptr==NULL) && (a->c_size)){
	  a->c_size = 0;
	  PyErr_NoMemory();
	}
    } else {
      if (PyObject_AsReadBuffer(v, &ptr, &size)) {
	  return -1;
      }
      if (size != 1) {
	  PyErr_SetString(PyExc_ValueError,
		  	  "can only assign buffer of size 1");
	  return -1;
      }
      *(char *) (a->c_ptr + i) = *(char *) ptr;
    }
    return 0;
}


static int
chunk_ass_slice(PyChunkObject * a, int start, int stop, PyObject * v)
{
    int size;
    const void *ptr;

    if (v!=NULL) {
      if (PyObject_AsReadBuffer(v, &ptr, &size)) {
	return -1;
      }
    } else {
      size = 0;
      ptr = NULL;
    }

    normalize_slice(a->c_size, &start, &stop);

    if (stop - start == size) {
	memmove(a->c_ptr + start, ptr, size);
    } else {
	/* we use a new buffer, this not the most efficient, but avoid
	   some chalanging situations like A[10:15]=A */
	void *nptr = malloc(a->c_size + size - (stop - start));
	if (nptr == NULL) {
	    PyErr_NoMemory();
	    return -1;
	}
	memcpy(nptr, a->c_ptr, start);
	memcpy(nptr + start, ptr, size);
	memcpy(nptr + start + size, a->c_ptr + stop, a->c_size - stop);
	if (a->c_ptr)
	    free(a->c_ptr);
	a->c_ptr = nptr;
	a->c_size += size - (stop - start);
    }
    return 0;
}


static int chunk_contains(PyChunkObject * self, PyObject * v)
{
    int size;
    const void *ptr;

    if (PyObject_AsReadBuffer(v, &ptr, &size)) {
	return -1;
    }

    /* optimize for a single character */
    if (size == 1) {
	return memchr(self->c_ptr, *(char *) ptr, self->c_size) != NULL;
    } else {
	char *lhs = self->c_ptr;
	char *end = lhs + self->c_size - size;
	while (lhs <= end) {
	    if (memcmp(lhs++, ptr, size) == 0)
		return 1;
	}
    }
    return 0;
}


static PyObject *chunk_inplace_concat(PyChunkObject * self, PyObject * bb)
{
    int size;
    const void *ptr;
    if (PyObject_AsReadBuffer(bb, &ptr, &size)) {
	return NULL;
    }
    if (chunk_do_resize(self, self->c_size + size, bb)) {
	return NULL;
    }
    Py_INCREF(self);
    return (PyObject *) self;
}

static PyObject *chunk_inplace_repeat(PyChunkObject * self, int n)
{
    if (n < 0)
	n = 0;
    if (chunk_do_resize(self, self->c_size * n, (PyObject *) self)) {
	return NULL;
    }
    Py_INCREF(self);
    return (PyObject *) self;
}

/* ------------------- buffer protocol ----------------------------*/

static int chunk_getbuffer(PyChunkObject * self, int idx, void **pp)
{
    if (idx != 0) {
	PyErr_SetString(PyExc_SystemError,
			"accessing non-existant buffer segment");
	return -1;
    }
    *pp = self->c_ptr;
    return self->c_size;
}
static int chunk_getsegcount(PyChunkObject * self, int *lenp)
{
    if (lenp) {
	*lenp = self->c_size;
    }
    return 1;
}
static int chunk_getcharbuf(PyChunkObject * self, int idx, const char **pp)
{
    if (idx != 0) {
	PyErr_SetString(PyExc_SystemError,
			"accessing non-existant buffer segment");
	return -1;
    }
    *pp = self->c_ptr;
    return self->c_size;
}

/* ------------------------ methods ----------------------------*/

PyDoc_STRVAR(resize__doc__, "c.resize(n, [pattern])\n\
\n\
Resize the memory segment to n bytes.\n\
If n is greater than the current segment size, the extra bytes are\n\
fill with pattern.");
static PyObject *chunk_resize(PyChunkObject * self, PyObject * args)
{
    int size;
    PyObject *data = NULL;

    if (!PyArg_ParseTuple(args, "i|O", &size, &data))
	return NULL;
    if (size < 0) {
	PyErr_SetString(PyExc_ValueError,
			"size argument must be positive");
	return NULL;
    }
    if (chunk_do_resize(self, size, data))
	return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

/* ------------------------ getter / setter ------------------------ */

static PyMethodDef chunk_methods[] = {
    /* instance methods */
    {"resize", (PyCFunction)chunk_resize, METH_VARARGS, resize__doc__},
    {"__reduce__",(PyCFunction)chunk_reduce, METH_NOARGS, reduce__doc__},
    {NULL, NULL}
};


static PyGetSetDef chunk_getsetlist[] = {
    {0}
};


static PySequenceMethods chunk_as_sequence = {
    (inquiry) chunk_length,	        /*sq_length */
    (binaryfunc) chunk_concat,      	/*sq_concat */
    (intargfunc) chunk_repeat,	        /*sq_repeat */
    (intargfunc) chunk_item,	        /*sq_item */
    (intintargfunc) chunk_slice,	/*sq_slice */
    (intobjargproc) chunk_ass_item,	/*sq_ass_item */
    (intintobjargproc) chunk_ass_slice,	/*sq_ass_slice */
    (objobjproc) chunk_contains,	/*sq_contains */
    (binaryfunc) chunk_inplace_concat,	/*sq_inplace_concat */
    (intargfunc) chunk_inplace_repeat	/*sq_inplace_repeat */
};

static PyBufferProcs chunk_as_buffer = {
    (getreadbufferproc) chunk_getbuffer,
    (getwritebufferproc) chunk_getbuffer,
    (getsegcountproc) chunk_getsegcount,
    (getcharbufferproc) chunk_getcharbuf,
};

PyDoc_STRVAR(chunk_type__doc__,"chunk([buffer [,buffer [,..]]]) --> chunk.\n\
\n\
Create a new chunk object filled by joining the argument buffers.\n");

static PyTypeObject PyChunk_Type = {
    PyObject_HEAD_INIT(&PyType_Type) 0,	/*ob_size */
    "mem.chunk.chunk",		/* tp_name */
    sizeof(PyChunkObject),	/* tp_basicsize */
    0,				/* tp_itemsize */
    (destructor) chunk_dealloc,	/* tp_dealloc */
    0,				/* tp_print */
    0,				/* tp_getattr */
    0,				/* tp_setattr */
    0,				/* tp_compare */
    (reprfunc) chunk_repr,	/* tp_repr */
    0,				/* tp_as_number */
    &chunk_as_sequence,		/* tp_as_sequence */
    0,				/* tp_as_mapping */
    0,				/* tp_hash */
    0,				/* tp_call */
    (reprfunc) chunk_str,	/* tp_str */
    0,				/* tp_getattro */
    0,				/* tp_setattro */
    &chunk_as_buffer,		/* tp_as_buffer */
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
      Py_TPFLAGS_HAVE_GETCHARBUFFER, /* tp_flags */
    chunk_type__doc__,		/* tp_doc */
    0,				/* tp_traverse */
    0,				/* tp_clear */
    chunk_richcompare,		/* tp_richcompare */
    0,				/* tp_weaklistoffset */
    0,				/* tp_iter */
    0,				/* tp_iternext */
    chunk_methods,		/* tp_methods */
    0,				/* tp_members */
    chunk_getsetlist,		/* tp_getset */
    0,				/* tp_base */
    0,				/* tp_dict */
    0,				/* tp_descr_get */
    0,				/* tp_descr_set */
    0,				/* tp_dictoffset */
    0,				/* tp_init */
    0,				/* tp_alloc */
    chunk_new,			/* tp_new */
    _PyObject_Del,		/* tp_free */
};

static PyMethodDef module_methods[] = {
    {NULL, NULL, 0, NULL}
};

void initchunk(void)
{
    PyObject *m;

    m = Py_InitModule3("chunk", module_methods, module__doc__);

    if (PyType_Ready(&PyChunk_Type) < 0)
	return;
    Py_INCREF(&PyChunk_Type);
    PyModule_AddObject(m, "chunk", (PyObject *) & PyChunk_Type);
}
