/* 
Copyright (c) 2003, Sebastien Keim, <s.keim@laposte.net>.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not
be used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization of the copyright holder.
*/

#include "Python.h"

typedef struct {
    PyObject_HEAD PyObject * b_base;
    int b_start;
    int b_stop;
} PyMemsliceObject;

extern PyTypeObject PyMemslice_Type;


PyDoc_STRVAR(module__doc__, "memory slices.\n\
\n\
This module define a new object types wich can be used to reference\n\
parts of object's memory buffer.");

/* private utilities */
static void
normalize_internal(PyMemsliceObject * self, const void **pp, int *size)
{
    int start = self->b_start;
    if (start < 0) {
	start += *size;
	if (start < 0)
	    start = 0;
    }
    if (start > *size) {
	start = *size;
	*size = 0;
    } else {
	if (self->b_stop >= 0) {
	    if (*size > (self->b_stop - start)) {
		*size = self->b_stop - start;
		if (*size < 0)
 		    *size = 0;
	    } else {
		*size -= start;
	    }
	} else {
	    *size -= start - self->b_stop;
	    if (*size < 0)
		*size = 0;
	}
    }
    *pp += start;
}

static PyObject *memslice_subtype_new(PyTypeObject * type, PyObject * args,
				      PyObject * kw);

static PyMemsliceObject *memslice_create(PyObject * base, int start,
					 int stop)
{
    PyMemsliceObject *buf;

    buf = PyObject_GC_New(PyMemsliceObject, &PyMemslice_Type);
    if (buf == NULL)
	return NULL;

    Py_INCREF(base);
    buf->b_base = base;
    buf->b_start = start;
    buf->b_stop = stop;

    PyObject_GC_Track((PyObject *) buf);

    return buf;
}

/* Methods */

static PyObject *memslice_new(PyTypeObject * type, PyObject * args,
			      PyObject * kw)
{
    PyObject *base;
    int start = 0;
    int stop = INT_MAX;

    static char* keywords[] = {"base", "start", "stop", NULL};

    if (type != &PyMemslice_Type)
	return memslice_subtype_new(type, args, kw);

    if (!PyArg_ParseTupleAndKeywords(args, kw, "O|ii", keywords,  &base, 
                                     &start, &stop))
	return NULL;


    if (!PyObject_CheckReadBuffer(base)) {
	PyErr_SetString(PyExc_ValueError,
			"data must support the buffer interface");
	return NULL;
    }

    return (PyObject *) memslice_create(base, start, stop);
}

static PyObject *memslice_subtype_new(PyTypeObject * type, PyObject * args,
				      PyObject * kw)
{
    PyMemsliceObject *tmp, *new;

    assert(PyType_IsSubtype(type, &PyMemslice_Type));
    tmp = (PyMemsliceObject *) memslice_new(&PyMemslice_Type, args, kw);
    if (tmp == NULL)
	return NULL;

    new = (PyMemsliceObject *) type->tp_alloc(type, 0);
    if (new == NULL)
	return NULL;

    new->b_base = tmp->b_base;
    new->b_start = tmp->b_start;
    new->b_stop = tmp->b_stop;

    Py_DECREF(tmp);
    return (PyObject *) new;
}

static void memslice_dealloc(PyMemsliceObject * self)
{

    Py_TRASHCAN_SAFE_BEGIN(self);
    Py_XDECREF(self->b_base);
    self->ob_type->tp_free((PyObject *) self);
    Py_TRASHCAN_SAFE_END(self);

}

static PyObject *memslice_repr(PyMemsliceObject * self)
{
    return PyString_FromFormat(
	 "<memslice start %d, stop %d at %p for object at %p>",
	 self->b_start, self->b_stop, self, self->b_base);
}

static PyObject *memslice_str(PyMemsliceObject * self)
{
    int size;
    const void *ptr;
    if (PyObject_AsReadBuffer(self->b_base, &ptr, &size)) {
	return NULL;
    }
    normalize_internal(self, &ptr, &size);
    return PyString_FromStringAndSize(ptr, size);
}

static long memslice_hash(PyMemsliceObject * self)
{
    /* If y remember well, C standard require:
     sizeof(int)==sizeof(void*) and sizeof(long)>=sizeof(int) */
    long x = (int)self->b_base + self->b_start + self->b_stop*28351; 

    if (x==-1)
        x = -2;
    return x;
}

static PyObject *memslice_richcompare(PyObject * va, PyObject * vb, int op)
{
    PyMemsliceObject *a = (PyMemsliceObject *) va;
    PyMemsliceObject *b = (PyMemsliceObject *) vb;
    int c;
    PyObject *result;

    /* Make sure both arguments are chunks. */
    if (!(PyObject_TypeCheck(a, &PyMemslice_Type) &&
	  PyObject_TypeCheck(b, &PyMemslice_Type))) {
	result = Py_NotImplemented;
	goto out;
    }

    if (a->b_base != b->b_base) {
	c = (a->b_base < b->b_base) ? -1 : 1;
    } else if (a->b_start != b->b_start) {
	c = (a->b_start < b->b_start) ? -1 : 1;
    } else if (a->b_stop != b->b_stop) {
	c = (a->b_stop < b->b_stop) ? -1 : 1;
    } else {
	c = 0;
    }

    switch (op) {
    case Py_LT:
	c = c < 0;
	break;
    case Py_LE:
	c = c <= 0;
	break;
    case Py_EQ:
	c = c == 0;
	break;
    case Py_NE:
	c = c != 0;
	break;
    case Py_GT:
	c = c > 0;
	break;
    case Py_GE:
	c = c >= 0;
	break;
    default:
	result = Py_NotImplemented;
	goto out;
    }
    result = c ? Py_True : Py_False;
  out:
    Py_INCREF(result);
    return result;
}

static int
memslice_traverse(PyMemsliceObject * self, visitproc visit, void *args)
{

    if (self->b_base == NULL) {
	return 0;
    }
    return visit(self->b_base, args);
}

/* ------------------------- Pickling support ------------------------- */
/* __getstate__ isn't exposed */
static PyObject* 
memslice_getstate(PyMemsliceObject * self)
{
    return Py_BuildValue("Oii", self->b_base, self->b_start, self->b_stop);
}

PyDoc_STRVAR(reduce__doc__, "b.__reduce__() -->  (cls, state)\n\
\n\
pickle and copy support.\n");
static PyObject* 
memslice_reduce(PyMemsliceObject * self)
{
    return Py_BuildValue("ON", self->ob_type, memslice_getstate(self));
}



/* --------------------------mapping methods -------------------------- */

static int memslice_length(PyMemsliceObject * self)
{
    int size;
    const void *pp;

    if (PyObject_AsReadBuffer(self->b_base, &pp, &size)) {
	return -1;
    }
    normalize_internal(self, &pp, &size);
    return size;
}

/* We use the mapping __getitem__ protocol because we doesn't
 * want the slice to be normalized (as do the tp_slice sequence slot).
 */
static PyObject* 
memslice_subscript(PyMemsliceObject * self, PySliceObject* slice)
{
  int start= 0;
  int stop = INT_MAX;
  
  if (!PySlice_Check(slice)) {
    PyErr_SetString(PyExc_TypeError, "unsuscribable object");
    return NULL;
  }
  if (slice->step!=Py_None) {
    PyErr_SetString(PyExc_TypeError, 
                  "memslice doesn't support extended slicing");
    return NULL;
  }
  if (slice->start!=Py_None) {
      if (PyInt_Check(slice->start)){
	  start = PyInt_AsLong(slice->start);
      } else {
	  PyErr_SetString(PyExc_TypeError, 
                  "memslice only support integer slice indices");
          return NULL;
      }
  }
  if (slice->stop!=Py_None) {
      if (PyInt_Check(slice->stop)){
	  stop = PyInt_AsLong(slice->stop);
      } else {
	  PyErr_SetString(PyExc_ValueError, 
                  "memslice only support integer slice indices");
          return NULL;
      }
  }
  return (PyObject *) memslice_create((PyObject *)self, start, stop);   
}


static PyObject *memslice_slice(PyMemsliceObject * a, int start, int stop)
{
    return (PyObject *) memslice_create((PyObject *) a, start, stop);
}


/* ------------------------ Buffer methods -------------------------- */
static int
memslice_getreadbuf(PyMemsliceObject * self, int idx, const void **pp)
{
    int size;
    if (PyObject_AsReadBuffer(self->b_base, pp, &size)) {
	return -1;
    }
    normalize_internal(self, pp, &size);
    return size;
}

static int
memslice_getwritebuf(PyMemsliceObject * self, int idx, void **pp)
{
    int size;
    if (PyObject_AsWriteBuffer(self->b_base, pp, &size)) {
	return -1;
    }
    normalize_internal(self, (const void **) pp, &size);
    return size;
}


static int memslice_getsegcount(PyMemsliceObject * self, int *lenp)
{
    if (lenp) {
	const void *p;
	if (PyObject_AsReadBuffer(self->b_base, &p, lenp)) {
	    return -1;
	}
	normalize_internal(self, &p, lenp);
    }
    return 1;
}

static int
memslice_getcharbuf(PyMemsliceObject * self, int idx, const char **pp)
{
    int size;
    if (PyObject_AsCharBuffer(self->b_base, pp, &size)) {
	return -1;
    }
    normalize_internal(self, (const void **) pp, &size);
    return size;
}

/* ------------------------ getter / setter ------------------------ */

#if PY_VERSION_HEX < 0x2030000
static PyObject *memslice_safeunpickle(PyMemsliceObject * self, void *closure)
{
    return PyInt_FromLong(1);
}
#endif

static PyObject *memslice_getbase(PyMemsliceObject * self, void *closure)
{
    if (self->b_base) {
	Py_INCREF(self->b_base);
	return self->b_base;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *memslice_getstart(PyMemsliceObject * self, void *closure)
{
    return PyInt_FromLong(self->b_start);
}

static PyObject *memslice_getstop(PyMemsliceObject * self, void *closure)
{
    return PyInt_FromLong(self->b_stop);
}


static PyMappingMethods memslice_as_mapping = {
    (inquiry) memslice_length,	     /*mp_length */
    (binaryfunc) memslice_subscript, /*mp_subscript*/
    0
};

static PyBufferProcs memslice_as_buffer = {
    (getreadbufferproc) memslice_getreadbuf,
    (getwritebufferproc) memslice_getwritebuf,
    (getsegcountproc) memslice_getsegcount,
    (getcharbufferproc) memslice_getcharbuf,
};

static PyMethodDef memslice_methods[] = {
    {"__reduce__", (PyCFunction)memslice_reduce, METH_NOARGS, reduce__doc__},
    {NULL, NULL}
};


static PyGetSetDef memslice_getsetlist[] = {
#if PY_VERSION_HEX < 0x2030000
    {"__safe_for_unpickling__", (getter)memslice_safeunpickle, NULL, NULL},
#endif
    {"base", (getter) memslice_getbase, NULL, NULL},
    {"start", (getter) memslice_getstart, NULL, NULL},
    {"stop", (getter) memslice_getstop, NULL, NULL},
    {0}
};

PyDoc_STRVAR(memslice_type__doc__,
             "memslice( base [,start [,stop]]) --> slice\n\
\n\
Return a reference to a memory buffer of the base object.\n\
start and stop arguments are optionals and are interpreted as \n\
in slice notation.\n\
The base object must support the single-segment buffer interface.");

PyTypeObject PyMemslice_Type = {
    PyObject_HEAD_INIT(&PyType_Type)
	0,
    "mem.memslice.memslice",
    sizeof(PyMemsliceObject),
    0,
    (destructor) memslice_dealloc,	/* tp_dealloc */
    0,				/* tp_print */
    0,				/* tp_getattr */
    0,				/* tp_setattr */
    0,				/* tp_compare */
    (reprfunc) memslice_repr,	/* tp_repr */
    0,				/* tp_as_number */
    0,	                        /* tp_as_sequence */
    &memslice_as_mapping,	/* tp_as_mapping */
    (hashfunc) memslice_hash,	/* tp_hash */
    0,				/* tp_call */
    (reprfunc) memslice_str,	/* tp_str */
    PyObject_GenericGetAttr,	/* tp_getattro */
    0,				/* tp_setattro */
    &memslice_as_buffer,	/* tp_as_buffer */
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC |
       Py_TPFLAGS_BASETYPE,     /* tp_flags */
    memslice_type__doc__,	/* tp_doc */
    (traverseproc) memslice_traverse,	/* tp_traverse */
    0,				/* tp_clear */
    memslice_richcompare,	/* tp_richcompare */
    0,				/* tp_weaklistoffset */
    0,				/* tp_iter */
    0,				/* tp_iternext */
    memslice_methods,		/* tp_methods */
    0,				/* tp_members */
    memslice_getsetlist,	/* tp_getset */
    0,				/* tp_base */
    0,				/* tp_dict */
    0,				/* tp_descr_get */
    0,				/* tp_descr_set */
    0,				/* tp_dictoffset */
    0,				/* tp_init */
    0,				/* tp_alloc */
    memslice_new,		/* tp_new */
    _PyObject_GC_Del,		/* tp_free */
};


static PyMethodDef module_methods[] = {
    {NULL, NULL, 0, NULL}
};

void initmemslice(void)
{
    PyObject *m;

    m = Py_InitModule3("memslice", module_methods, module__doc__);

    if (PyType_Ready(&PyMemslice_Type) < 0)
	return;
    Py_INCREF(&PyMemslice_Type);
    PyModule_AddObject(m, "memslice", (PyObject *) & PyMemslice_Type);
}
