/* 
Copyright (c) 2003, Sebastien Keim, <s.keim@laposte.net>.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any purpose
with or without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not
be used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization of the copyright holder.
*/

#include "Python.h"


PyDoc_STRVAR(module__doc__, "Objects internal memory access.\n\
\n\
This module provide a set of low level functions to access internal \n\
memory of objects which define the single-segment buffer interface.");


PyDoc_STRVAR(memcopy__doc__, "memcopy(dest, src, [start=0])\n\
\n\
Copy src buffer into dest buffer at start location.\n\
Raise IndexError if src is too long for dest buffer.");
static PyObject *tools_memcopy(PyObject * self, PyObject * args)
{
    PyObject *src;
    PyObject *dest;
    const void *src_ptr;
    int src_len;
    void *dst_ptr;
    int start = 0;
    int dst_len;

    if (!PyArg_ParseTuple(args, "OO|i", &dest, &src, &start)) {
	return NULL;
    }
    if (PyObject_AsReadBuffer(src, &src_ptr, &src_len)) {
	PyErr_SetString(PyExc_TypeError,
			"source argument must support the buffer interface");
	return NULL;
    }
    if (PyObject_AsWriteBuffer(dest, &dst_ptr, &dst_len)) {
	PyErr_SetString(PyExc_TypeError,
			"dest argument must support the buffer interface");
	return NULL;
    }
    if (start < 0 || start > dst_len) {
	PyErr_SetString(PyExc_IndexError, "invalide index argument.");
	return NULL;
    }
    if (src_len + start > dst_len) {
	PyErr_SetString(PyExc_IndexError,
			"source argument too long for destination buffer");
	return NULL;
    }
    memmove(dst_ptr + start, src_ptr, src_len);
    Py_INCREF(Py_None);
    return Py_None;
}

/*
 *   ------------- index functions --------------------
 */
static void mem_adjust_indices(int *start, int *end, int len)
{
    if (*end > len)
	*end = len;
    else if (*end < 0)
	*end += len;
    if (*end < 0)
	*end = 0;
    if (*start < 0)
	*start += len;
    if (*start < 0)
	*start = 0;
}

static long mem_find_internal(PyObject * self, PyObject * args, int dir)
{
    PyObject *src;
    const char *src_ptr;
    int src_len;
    PyObject *mtf;
    const char *mtf_ptr;
    int mtf_len;
    int start = 0;
    int stop = INT_MAX;
    int step = 1;

    if (!PyArg_ParseTuple
	(args, "OO|iii", &src, &mtf, &start, &stop, &step)) {
	return -2;
    }
    if (PyObject_AsReadBuffer(src, (const void **) &src_ptr, &src_len)) {
	PyErr_SetString(PyExc_TypeError,
			"source argument must support the buffer interface");
	return -2;
    }
    if (PyObject_AsReadBuffer(mtf, (const void **) &mtf_ptr, &mtf_len)) {
	PyErr_SetString(PyExc_TypeError,
			"pattern argument must support the buffer interface");
	return -2;
    }

    mem_adjust_indices(&start, &stop, src_len);

    if (dir > 0) {
	if (mtf_len == 0 && start <= stop)
	    return (long) start;
	stop -= mtf_len;
	for (; start <= stop; start += step)
	    if (src_ptr[start] == mtf_ptr[0]
		&& memcmp(&src_ptr[start], mtf_ptr, mtf_len) == 0)
		return (long) start;
    } else {
	int j;
	if (mtf_len == 0 && start <= stop)
	    return (long) stop;
	for (j = stop - mtf_len; j >= start; j -= step)
	    if (src_ptr[j] == mtf_ptr[0]
		&& memcmp(&src_ptr[j], mtf_ptr, mtf_len) == 0)
		return (long) j;
    }
    return -1;
}

PyDoc_STRVAR(index__doc__,
	     "index(dest, pattern, [,start [,stop [,step]]]) --> int\n\
\n\
Return the lowest index in dest where pattern is found,\n\
such as the pattern  is contained in the range [start, end].\n\
Optional arguments start and end are interpreted as in slice notation\n\
The third argument step can be used to specify alignment.\n\
Raise a ValueError if pattern is not found.");
static PyObject *tools_index(PyObject * self, PyObject * args)
{
    long result = mem_find_internal(self, args, +1);
    if (result == -2)
	return NULL;
    if (result == -1) {
	PyErr_SetString(PyExc_ValueError, "pattern not found");
	return NULL;
    }
    return PyInt_FromLong(result);
}

PyDoc_STRVAR(rindex__doc__,
	     "rindex(dest, pattern, [,start [,stop [,step]]]) --> int\n\
\n\
Like index but return the hightest index.");
static PyObject *tools_rindex(PyObject * self, PyObject * args)
{
    long result = mem_find_internal(self, args, -1);
    if (result == -2)
	return NULL;
    if (result == -1) {
	PyErr_SetString(PyExc_ValueError, "pattern not found");
	return NULL;
    }
    return PyInt_FromLong(result);
}

PyDoc_STRVAR(memcmp__doc__, "memcmp(x, y) --> int\n\
\n\
Raw memory comparison\n\
Return negative if x<y, zero if x==y, positive if x>y.'");
static PyObject *tools_memcmp(PyObject * self, PyObject * args)
{
    PyObject *a;
    PyObject *b;
    int result = 0;
    if (!PyArg_ParseTuple(args, "OO", &a, &b)) {
	return NULL;
    }
    if (a != b) {
	const void *ptr_a, *ptr_b;
	int size_a, size_b;
	int min_size;
	if (PyObject_AsReadBuffer(a, &ptr_a, &size_a)) {
	    return NULL;
	}
	if (PyObject_AsReadBuffer(b, &ptr_b, &size_b)) {
	    return NULL;
	}
	min_size = (size_a < size_b) ? size_a : size_b;
	result = memcmp(ptr_a, ptr_b, min_size);
	if (result == 0) {
	    if (size_a != size_b) {
		result = (size_a < size_b) ? -1 : 1;
	    }
	} else {
	    result = (result < 0) ? -1 : 1;
	}
    }
    return PyInt_FromLong(result);
}

PyDoc_STRVAR(memlen__doc__, "memlen(x) --> int\n\
\n\
Return the length in bytes of object's buffer.");
static PyObject *tools_memlen(PyObject * self, PyObject * args)
{
    PyObject *a;
    int size;
    const void *ptr;

    if (!PyArg_ParseTuple(args, "O", &a)) {
	return NULL;
    }
    if (PyObject_AsReadBuffer(a, &ptr, &size)) {
	return NULL;
    }

    return PyInt_FromLong(size);
}

PyObject* tools_match_internal(PyObject* args, int dir, int equal)
{
    PyObject *a, *b;
    int size_a, size_b;
    const void *ptr_a, *ptr_b;
    char *x, *y, *end;

    if (!PyArg_ParseTuple(args, "OO:match/diff/rmatch/rdiff", &a, &b)) {
	return NULL;
    }
    if (PyObject_AsReadBuffer(a, &ptr_a, &size_a)) {
	return NULL;
    }
    if (PyObject_AsReadBuffer(b, &ptr_b, &size_b)) {
	return NULL;
    }
     
    if (dir>0) {
         x = (char*)ptr_a;
	 y = (char*)ptr_b;
	 end = (char*)ptr_a + (size_a<size_b ? size_a : size_b);
    } else {
         x = (char*)ptr_a + size_a - 1;
	 y = (char*)ptr_b + size_b - 1;
	 end = x - (size_a<size_b ? size_a : size_b);
    }
    
    for (; x!=end; x+=dir, y+=dir) {
      if ((*x==*y)==equal) {
	return PyInt_FromLong(x - (char*)ptr_a);
      }
    }
    return PyInt_FromLong(-1);
}


PyDoc_STRVAR(match__doc__, "match(x, y) --> int\n\
\n\
Return the lowest index where x and y have a common byte.\n\
Return -1 if the two buffers have no bytes in common.");
static PyObject *tools_match(PyObject * self, PyObject * args)
{
  return tools_match_internal(args, 1, 1); 
}
PyDoc_STRVAR(rmatch__doc__, "rmatch(x, y) --> int\n\
\n\
Same as match() but return the hightest index.\n\
The returned index corresponds to the x buffer.");
static PyObject *tools_rmatch(PyObject * self, PyObject * args)
{
  return tools_match_internal(args, -1, 1); 
}

PyDoc_STRVAR(diff__doc__, "diff(x, y) --> int\n\
\n\
Return the lowest index where x and y have a different byte.\n\
Return -1 if the two buffers have all bytes in common.");
static PyObject *tools_diff(PyObject * self, PyObject * args)
{
  return tools_match_internal(args, 1, 0); 
}

PyDoc_STRVAR(rdiff__doc__, "rdiff(x, y) --> int\n\
\n\
Same as diff() but return the hightest index.\n\
The returned index corresponds to the x buffer.");
static PyObject *tools_rdiff(PyObject * self, PyObject * args)
{
  return tools_match_internal(args, -1, 0); 
}


static PyMethodDef module_methods[] = {
    {"memcopy", tools_memcopy, METH_VARARGS, memcopy__doc__},
    {"index", tools_index, METH_VARARGS, index__doc__},
    {"rindex", tools_rindex, METH_VARARGS, rindex__doc__},
    {"memcmp", tools_memcmp, METH_VARARGS, memcmp__doc__},
    {"memlen", tools_memlen, METH_VARARGS, memlen__doc__},
    {"match", tools_match, METH_VARARGS, match__doc__},
    {"rmatch", tools_rmatch, METH_VARARGS, rmatch__doc__},
    {"diff", tools_diff, METH_VARARGS, diff__doc__},
    {"rdiff", tools_rdiff, METH_VARARGS, rdiff__doc__},
    {NULL, NULL, 0, NULL}
};

void inittools(void)
{
    PyObject *m;
    m = Py_InitModule3("tools", module_methods, module__doc__);
}
