"Test SubBuffer object"


import unittest
from mem.memslice import memslice
from mem.chunk import chunk
from mem.tools import memcopy

import pickle
import gc

class MemsliceTester(unittest.TestCase):

    def test_create(self):
        x = memslice("hello")
        self.assertEqual(str(x), "hello")
        x = memslice("hello", 2)
        self.assertEqual(str(x), "llo")
        x = memslice("hello", 2, 4)
        self.assertEqual(str(x), "ll")
        x = memslice("hello", -2)
        self.assertEqual(str(x), "lo")
        x = memslice("hello", -3, -1)
        self.assertEqual(str(x), "ll")
        x = memslice("hello", -30, 2)
        self.assertEqual(str(x), "he")
        x = memslice(**{"base":"hello", "start":-3, "stop":20})
        self.assertEqual(str(x), "llo")

        str(memslice("This cause a segfault on v0.1", -4, 1))

        #check for bug in 0.2: 0 for end index was interperpreted as
        #a negative indice (counted from end)
        a = memslice("Hello",0,0)
        self.assertEqual(len(a), 0)


        
    def test_resize(self):
        c = chunk(5* "*")
        x = memslice(c,0,5)
        self.assertEqual(str(x), "*****")
        c.resize(1)
        self.assertEqual(str(x), "*")
        c.resize(100, '*')
        self.assertEqual(str(x), "*****")

    def test_mutable(self):
        c = chunk(5*"*")
        x = memslice(c,1)
        memcopy(x, "+++")
        self.assertEqual(str(c), "*+++*")
        memcopy(x, "*",1)
        self.assertEqual(str(c), "*+*+*")

    def test_non_mutable(self):
        x = memslice("hello")
        self.assertRaises(TypeError, memcopy, x, "*")

    def test_compare(self):
        x = "hello"
        # memslice are compared:
        #  1 - by the address of the base object
        #  2 - by the index
        #  3 - by the size
        self.assert_(memslice("hello") != memslice("*hello" ,1))
        self.assert_(memslice(x) == memslice(x))
        self.assert_(memslice(x,1) > memslice(x))
        self.assert_(memslice(x,1,4) > memslice(x,1,2))

    def test_hash(self):
        x = chunk("hello")
        a = memslice(x); A = hash(a)
        b = memslice(x); B = hash(b)
        c = memslice(x, 1, -1); C = hash(c)
        dic = {a:None}

        self.assertEqual(A, hash(b))
        # check that common collide are avoided
        self.failIfEqual(A, hash(c))

        #check that target mutations doesn't change hash value
        x.resize(100)
        x[0] = 'H'
        self.assertEqual(A, hash(a))
        self.assertEqual(B, hash(b))
        self.assertEqual(C, hash(c))
        self.assertEqual(dic[b], None)

    def test_gc(self):
        class A(str):
            pass
        a = A()
        a.x = memslice(a)
        gc.collect()
        del a
        self.assert_(gc.collect()>0, "non collectable cycle")

    def test_attributes(self):
        s = "a string buffer"
        a = memslice(s,1,10)
        self.assert_(a.base is s)
        self.assertEqual(a.start, 1)
        self.assertEqual(a.stop, 10)

    def test_subclass(self):
        class A(memslice):
            def __new__(cls, data, start, stop):
                obj = memslice.__new__(cls, data, start, stop)
                obj.limits = (start, stop)
                return obj
        s = "Hello"
        a = A(s,1,-1)
        self.assertEqual(str(a),"ell")
        self.assertEqual(a.limits, (1, -1))

    def test_slice(self):
        a = memslice("hello", 1)
        b = a[1:]
        c = a[-1:]
        self.assertEqual(str(b), "llo")
        self.assertEqual(str(c), "o")

        # check for bug in version 0.1: start and stop where
        # normalized by slicing
        buf = chunk("Hello World")
        a = memslice(buf)
        b = a[-5:-1]
        self.assertEqual(str(b), "Worl")
        buf+="*"
        self.assertEqual(str(b), "orld")

        # now that we use mp_subscript, check that we catch bad keys
        a = memslice("1234")
        self.assertRaises(TypeError, a.__getitem__, None)
        self.assertRaises(TypeError, a.__getitem__, 1)
        self.assertRaises(TypeError, a.__getitem__, slice(1,2,3))
        self.assertRaises(TypeError, a.__getitem__, slice("1",2))
        
        

    def test_pickle(self):
        buf1 = "Hello World"
        a1 = memslice(buf1, 1, -1)

        s = pickle.dumps((a1, buf1))
        a2, buf2 = pickle.loads(s)
        # we can't test for equality, because it's not the
        # same base object
        self.assertEqual(str(buf1), str(buf2))
        self.assert_(buf1 is not buf2)
        # but the pickle protocol has kept the relation between
        # the memslice and it's base object:
        self.assert_(a2.base is  buf2)
        

from test.test_support import run_suite
def test_main():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemsliceTester))
    run_suite(suite)

if __name__ == '__main__':
    test_main()
